import tensorflow as tf
from tensorflow.python.keras.metrics import Mean
from tensorflow.python.keras import backend as K

# for future reference:
# B: batch size
# M: intent number
# n: where n is the tag set number
# n+2: n plus start and end tag
# N: n * n
# T: sequence length
# F: feature number


class SequenceSpanAccuracy(Mean):
    def __init__(self, name="sequence_span_accuracy", dtype=None):
        """Creates a `CategoricalAccuracy` instance.
        Args:
          name: (Optional) string name of the metric instance.
          dtype: (Optional) data type of the metric result.
        """
        super(SequenceSpanAccuracy, self).__init__(
            name=name,
            dtype=dtype
        )

    def update_state(self, y_true, y_pred, *args, **kwargs):
        """

        :param y_true: Tensor shape: (B, T)
        :param y_pred: Tensor shape (B, T)
        :param sample_weight: None or Tensor shape (B, T)
        :return:
        """
        value = K.equal(y_pred, y_true)  # shape: (B, T)

        # # # DEBUG: output training value
        # print_op = tf.print(mask)
        # with tf.control_dependencies([print_op]):
        #     value = tf.identity(value)

        super(SequenceSpanAccuracy, self).update_state(value, *args, **kwargs)


def get_mask_from_keras_tensor(y_pred):
    layer, node_index = y_pred._keras_history[:2]
    mask = layer.output_mask

    return mask


def sequence_span_accuracy(y_true, y_pred):
    """

    :param y_true: Tensor shape: (B, T)
    :param y_pred: Tensor shape (B, T)
    :return:
    """
    judge = K.cast(K.equal(y_pred, y_true), K.floatx())  # shape: (B, T)
    mask = get_mask_from_keras_tensor(y_pred)  # shape: (B, T)
    if mask is None:
        result = K.mean(judge)
        return result
    else:
        mask = K.cast(mask, K.floatx())
        result = K.sum(judge * mask) / K.sum(mask)
        return result



"""
Conditional random field
"""
from collections import namedtuple
from typing import List, Tuple, Dict


from tf_crf_layer.exceptions import ConfigurationError

ConstraintType = namedtuple('ConstraintType', ["BIO", "IOB1", "BIOUL", "BMES"])
constraint_type = ConstraintType("BIO", "IOB1", "BIOUL", "BMES")


def allowed_transitions(constraint_type: str, labels: Dict[int, str]) -> List[Tuple[int, int]]:
    """
    Given labels and a constraint type, returns the allowed transitions. It will
    additionally include transitions for the start and end states, which are used
    by the conditional random field.

    Parameters
    ----------
    constraint_type : ``str``, required
        Indicates which constraint to apply. Current choices are
        "BIO", "IOB1", "BIOUL", and "BMES".
    labels : ``Dict[int, str]``, required
        A mapping {label_id -> label}. Most commonly this would be the value from
        Vocabulary.get_index_to_token_vocabulary()

    Returns
    -------
    ``List[Tuple[int, int]]``
        The allowed transitions (from_label_id, to_label_id).
    """
    num_labels = len(labels)
    start_tag = num_labels
    end_tag = num_labels + 1
    labels_with_boundaries = list(labels.items()) + [(start_tag, "START"), (end_tag, "END")]

    allowed = []
    for from_label_index, from_label in labels_with_boundaries:
        if from_label in ("START", "END"):
            from_tag = from_label
            from_entity = ""
        else:
            from_tag = from_label[0]
            from_entity = from_label[1:]
        for to_label_index, to_label in labels_with_boundaries:
            if to_label in ("START", "END"):
                to_tag = to_label
                to_entity = ""
            else:
                to_tag = to_label[0]
                to_entity = to_label[1:]
            if is_transition_allowed(constraint_type, from_tag, from_entity,
                                     to_tag, to_entity):
                allowed.append((from_label_index, to_label_index))
    return allowed


def is_transition_allowed(constraint_type: str,
                          from_tag: str,
                          from_entity: str,
                          to_tag: str,
                          to_entity: str):
    """
    Given a constraint type and strings ``from_tag`` and ``to_tag`` that
    represent the origin and destination of the transition, return whether
    the transition is allowed under the given constraint type.

    Parameters
    ----------
    constraint_type : ``str``, required
        Indicates which constraint to apply. Current choices are
        "BIO", "IOB1", "BIOUL", and "BMES".
    from_tag : ``str``, required
        The tag that the transition originates from. For example, if the
        label is ``I-PER``, the ``from_tag`` is ``I``.
    from_entity: ``str``, required
        The entity corresponding to the ``from_tag``. For example, if the
        label is ``I-PER``, the ``from_entity`` is ``PER``.
    to_tag : ``str``, required
        The tag that the transition leads to. For example, if the
        label is ``I-PER``, the ``to_tag`` is ``I``.
    to_entity: ``str``, required
        The entity corresponding to the ``to_tag``. For example, if the
        label is ``I-PER``, the ``to_entity`` is ``PER``.

    Returns
    -------
    ``bool``
        Whether the transition is allowed under the given ``constraint_type``.
    """
    # pylint: disable=too-many-return-statements
    if to_tag == "START" or from_tag == "END":
        # Cannot transition into START or from END
        return False

    if constraint_type == "BIOUL":
        if from_tag == "START":
            return to_tag in ('O', 'B', 'U')
        if to_tag == "END":
            return from_tag in ('O', 'L', 'U')
        return any([
                # O can transition to O, B-* or U-*
                # L-x can transition to O, B-*, or U-*
                # U-x can transition to O, B-*, or U-*
                from_tag in ('O', 'L', 'U') and to_tag in ('O', 'B', 'U'),
                # B-x can only transition to I-x or L-x
                # I-x can only transition to I-x or L-x
                from_tag in ('B', 'I') and to_tag in ('I', 'L') and from_entity == to_entity
        ])
    elif constraint_type == "BIO":
        if from_tag == "START":
            return to_tag in ('O', 'B')
        if to_tag == "END":
            return from_tag in ('O', 'B', 'I')
        return any([
                # Can always transition to O or B-x
                to_tag in ('O', 'B'),
                # Can only transition to I-x from B-x or I-x
                to_tag == 'I' and from_tag in ('B', 'I') and from_entity == to_entity
        ])
    elif constraint_type == "IOB1":
        if from_tag == "START":
            return to_tag in ('O', 'I')
        if to_tag == "END":
            return from_tag in ('O', 'B', 'I')
        return any([
                # Can always transition to O or I-x
                to_tag in ('O', 'I'),
                # Can only transition to B-x from B-x or I-x, where
                # x is the same tag.
                to_tag == 'B' and from_tag in ('B', 'I') and from_entity == to_entity
        ])
    elif constraint_type == "BMES":
        if from_tag == "START":
            return to_tag in ('B', 'S')
        if to_tag == "END":
            return from_tag in ('E', 'S')
        return any([
                # Can only transition to B or S from E or S.
                to_tag in ('B', 'S') and from_tag in ('E', 'S'),
                # Can only transition to M-x from B-x, where
                # x is the same tag.
                to_tag == 'M' and from_tag in ('B', 'M') and from_entity == to_entity,
                # Can only transition to E-x from B-x or M-x, where
                # x is the same tag.
                to_tag == 'E' and from_tag in ('B', 'M') and from_entity == to_entity,
        ])
    else:
        raise ConfigurationError(f"Unknown constraint type: {constraint_type}")


import tensorflow as tf


def _get_accuracy(y_true, y_pred, mask, sparse_target=False):
    y_pred = tf.keras.backend.argmax(y_pred, -1)
    if sparse_target:
        y_true = tf.keras.backend.cast(y_true[:, :, 0], tf.keras.backend.dtype(y_pred))
    else:
        y_true = tf.keras.backend.argmax(y_true, -1)
    judge = tf.keras.backend.cast(tf.keras.backend.equal(y_pred, y_true), tf.keras.backend.floatx())
    if mask is None:
        return tf.keras.backend.mean(judge)
    else:
        mask = tf.keras.backend.cast(mask, tf.keras.backend.floatx())
        return tf.keras.backend.sum(judge * mask) / tf.keras.backend.sum(mask)


def crf_viterbi_accuracy(y_true, y_pred):
    """
    Use Viterbi algorithm to get best path, and compute its accuracy.
    `y_pred` must be an output from CRF.
    """
    crf, idx = y_pred._keras_history[:2]
    # X = crf._inbound_nodes[idx].input_tensors[0]
    # mask = crf._inbound_nodes[idx].input_masks[0]
    # y_pred = crf.viterbi_decoding(X, mask)
    # return _get_accuracy(y_true, y_pred, mask, crf.sparse_target)
    return crf.get_accuracy(y_true, y_pred)


def crf_marginal_accuracy(y_true, y_pred):
    """
    Use time-wise marginal argmax as prediction.
    `y_pred` must be an output from CRF with `learn_mode="marginal"`.
    """
    crf, idx = y_pred._keras_history[:2]
    X = crf._inbound_nodes[idx].input_tensors[0]
    mask = crf._inbound_nodes[idx].input_masks[0]
    y_pred = crf.get_marginal_prob(X, mask)
    return _get_accuracy(y_true, y_pred, mask, crf.sparse_target)


def crf_accuracy(y_true, y_pred):
    """
    Ge default accuracy based on CRF `test_mode`.
    """
    crf, idx = y_pred._keras_history[:2]
    if crf.test_mode == 'viterbi':
        return crf_viterbi_accuracy(y_true, y_pred)
    else:
        return crf_marginal_accuracy(y_true, y_pred)



from tf_crf_layer.metrics.crf_accuracy import crf_accuracy, crf_viterbi_accuracy
from tf_crf_layer.metrics.sequence_correctness import SequenceCorrectness
from tf_crf_layer.metrics.sequence_span_accuracy import SequenceSpanAccuracy, sequence_span_accuracy



import itertools
import json

import numpy as np
import pandas as pd

from tf_crf_layer.crf_static_constraint_helper import allowed_transitions
from tokenizer_tools.tagset.NER.BILUO import BILUOEncoderDecoder

from typing import Sequence, Tuple, List, AnyStr, Dict


def generate_constraint_table(constraint_mapping: List[List[str]] = None, tag_dict: Dict[int, str] = None, tag_scheme='BIOUL'):
    constraint_matrix_list = []
    for constraint in constraint_mapping:
        all_tag_set = set()

        for entity in constraint:
            encoder = BILUOEncoderDecoder(entity)
            tag_set = encoder.all_tag_set()

            all_tag_set.update(tag_set)

        all_tag_dict = dict(i for i in tag_dict.items() if i[1] in all_tag_set)

        _, constraint_matrix = allowed_transitions(tag_scheme, all_tag_dict, tag_dict)

        constraint_matrix_list.append(constraint_matrix)

    constraint_matrix = np.stack([df.values for df in constraint_matrix_list], axis=0)
    return constraint_matrix


def filter_constraint(constraint: Dict[str, List[str]], tag_set: List[str] = None, intent_set: List[str] = None):
    """
    filter out entity not in tag_list

    :param constraint:
    :param tag_set:
    :param intent_set:
    :return:
    """

    valid_constraint = dict()
    for k, v in constraint.items():
        if intent_set and k not in intent_set:
            # discard
            continue

        # filter out tag that not in tag set
        if not tag_set:
            valid_v = v
        else:
            valid_v = list(filter(lambda x: x in tag_set, v))

        valid_constraint[k] = valid_v

    return valid_constraint


def sort_constraint(constraints: Dict[str, List[str]], intent_lookup_table: Dict[str, int]):
    sorted_constraint = sorted(constraints.items(), key=lambda x: intent_lookup_table[x[0]])
    sorted_constraint_list = [i[1] for i in sorted_constraint]

    return sorted_constraint_list
        

if __name__ == "__main__":
    with open('/home/howl/PycharmProjects/seq2annotation/data/constraint.json') as fd:
        constraint = json.load(fd)
        
    with open('/home/howl/PycharmProjects/seq2annotation/data/entity.txt') as fd:
        tag_list = [i.strip() for i in fd]
        tag_set_list = [BILUOEncoderDecoder(i).all_tag_set() for i in tag_list]

        tag_set = set()
        for tag in tag_set_list:
            tag_set.update(set(tag))

    # filter out entity not in tag_list
    valid_constraint = filter_constraint(constraint, tag_list)

    constraint_mapping = list(valid_constraint.values())
    tag_dict = dict(enumerate(tag_set))
        
    generate_constraint_table(constraint_mapping, tag_dict)




import itertools
import math

import numpy as np


class CRF(object):
    def __init__(self, units, transitions, transitions_from_start, transitions_to_end, transition_constraint_mask=None, dynamic_transition_constraint=None):
        self.units = units
        self.transitions = transitions
        self.transitions_from_start = transitions_from_start
        self.transitions_to_end = transitions_to_end
        self.transition_constraint_mask = transition_constraint_mask
        self.dynamic_transition_constraint = dynamic_transition_constraint

    def score(self, logits, tags, transitions_from_start=None, transitions_to_end=None, transitions=None):
        """
        Computes the likelihood score for the given sequence of tags,
        given the provided logits (and the transition weights in the CRF model)
        """
        if transitions_from_start is None:
            transitions_from_start = self.transitions_from_start

        if transitions_to_end is None:
            transitions_to_end = self.transitions_to_end

        if transitions is None:
            transitions = self.transitions

        # Start with transitions from START and to END
        total = transitions_from_start[tags[0]] + transitions_to_end[tags[-1]]
        # Add in all the intermediate transitions
        for tag, next_tag in zip(tags, tags[1:]):
            total += transitions[tag, next_tag]
        # Add in the logits for the observed tags
        for logit, tag in zip(logits, tags):
            total += logit[tag]
        return total

    def compute_log_likelihood_without_mask(self, logits, tags):
        # Now compute the log-likelihood manually
        manual_log_likelihood = 0.0

        # For each instance, manually compute the numerator
        # (which is just the score for the logits and actual tags)
        # and the denominator
        # (which is the log-sum-exp of the scores for the logits across all possible tags)
        for logits_i, tags_i in zip(logits, tags):
            numerator = self.score(logits_i, tags_i)
            all_scores = [self.score(logits_i, tags_j)
                          for tags_j in itertools.product(range(5), repeat=3)]
            denominator = math.log(sum(math.exp(score) for score in all_scores))
            # And include them in the manual calculation.
            manual_log_likelihood += numerator - denominator

        return manual_log_likelihood

    def compute_log_likelihood_with_mask(self, logits, tags, mask):
        # Now compute the log-likelihood manually
        manual_log_likelihood = 0.0

        # For each instance, manually compute the numerator
        #   (which is just the score for the logits and actual tags)
        # and the denominator
        #   (which is the log-sum-exp of the scores for the logits across all possible tags)
        for logits_i, tags_i, mask_i in zip(logits, tags, mask):
            # Find the sequence length for this input and only look at that much of each sequence.
            sequence_length = np.sum(mask_i)
            logits_i = logits_i[:sequence_length]
            tags_i = tags_i[:sequence_length]

            numerator = self.score(logits_i, tags_i)
            all_scores = [self.score(logits_i, tags_j)
                          for tags_j in itertools.product(range(5), repeat=sequence_length)]
            denominator = math.log(sum(math.exp(score) for score in all_scores))
            # And include them in the manual calculation.
            manual_log_likelihood += numerator - denominator

        return manual_log_likelihood

    def compute_effective_boundary(self):
        """

        :return: Tensor(shape=(B, 1, n)) or Tensor(shape=(1, 1, n))
        """
        left_boundary = self.transitions_from_start
        right_boundary = self.transitions_to_end

        if self.transition_constraint_mask is not None:
            start_tag = self.units
            end_tag = self.units + 1

            left_boundary = (
                    self.transitions_from_start * self.transition_constraint_mask[start_tag, :self.units] +
                    -10000.0 * (1 - self.transition_constraint_mask[start_tag, :self.units])
            )
            right_boundary = (
                    self.transitions_to_end * self.transition_constraint_mask[self.units, end_tag] +
                    -10000.0 * (1 - self.transition_constraint_mask[self.units, end_tag])
            )

        if self.dynamic_transition_constraint is not None:
            start_tag = self.units
            end_tag = self.units + 1

            # shape: (B, 1, n)
            left_dynamic_constraint = self.dynamic_transition_constraint[:, start_tag, :self.units]

            # shape: (B, 1, n)
            left_boundary = (
                self.transitions_from_start * left_dynamic_constraint +
                -10000.0 * (1 - left_dynamic_constraint)
            )

            # shape: (B, 1, n)
            right_dynamic_constraint = self.dynamic_transition_constraint[:, :self.units, end_tag]

            # shape: (B, 1, n)
            right_boundary = (
                self.transitions_to_end * right_dynamic_constraint +
                -10000.0 * (1 - right_dynamic_constraint)
            )

        return left_boundary, right_boundary

    def compute_effective_chain_kernel(self):
        """

        :return: Tensor(shape=(B, n, n)) or Tensor(shape=(n, n))
        """
        chain_kernel = self.transitions

        if self.transition_constraint_mask is not None:
            chain_kernel_mask = self.transition_constraint_mask[:self.units, :self.units]
            chain_kernel = self.transitions * chain_kernel_mask + -10000.0 * (1 - chain_kernel_mask)

        if self.dynamic_transition_constraint is not None:
            dynamic_chain_kernel_mask = self.dynamic_transition_constraint[:, :self.units, :self.units]
            chain_kernel = self.transitions * dynamic_chain_kernel_mask + -10000.0 * (1 - dynamic_chain_kernel_mask)

        return chain_kernel

    def viterbi_decode(self, logits, mask=None):
        if mask is None:
            mask = np.ones(np.shape(logits)[:2], dtype=np.bool)

        transitions_from_start, transitions_to_end = self.compute_effective_boundary()
        transitions = self.compute_effective_chain_kernel()

        most_likely_tags = []
        best_scores = []

        for logit, mas, from_start, to_end, transition in zip(logits, mask, transitions_from_start, transitions_to_end, transitions):
            sequence_length = np.sum(mas)
            most_likely, most_likelihood = None, -float('inf')
            for tags in itertools.product(range(5), repeat=sequence_length):
                score = self.score(logit, tags, from_start, to_end, transition)
                if score > most_likelihood:
                    # padding tags to sequence length
                    tag_len_diff = 3 - len(tags)
                    if tag_len_diff:
                        tags = list(tags) + [0] * tag_len_diff
                        
                    most_likely, most_likelihood = tags, score
            # Convert tuple to list; otherwise == complains.
            most_likely_tags.append(list(most_likely))
            best_scores.append(most_likelihood)

        return most_likely_tags



"""
Conditional random field

Some code are copied from AllenNLP
"""
from collections import namedtuple
from typing import List, Tuple, Dict

import numpy as np
import pandas as pd

from tf_crf_layer.exceptions import ConfigurationError

ConstraintType = namedtuple('ConstraintType', ["BIO", "IOB1", "BIOUL", "BMES"])
constraint_type = ConstraintType("BIO", "IOB1", "BIOUL", "BMES")


def extract_tag_entity(label: str) -> (str, str):
    # extract tag and entity from label
    if label in ("START", "END"):
        tag = label
        entity = ""
    else:
        tag = label[0]
        entity = label[1:]

    return tag, entity


# def allowed_transitions(constraint_type: str, labels: Dict[int, str]) -> (List[Tuple[int, int]], pd.DataFrame):
#     """
#     Given labels and a constraint type, returns the allowed transitions. It will
#     additionally include transitions for the start and end states, which are used
#     by the conditional random field.
#
#     Parameters
#     ----------
#     constraint_type : ``str``, required
#         Indicates which constraint to apply. Current choices are
#         "BIO", "IOB1", "BIOUL", and "BMES".
#     labels : ``Dict[int, str]``, required
#         A mapping {label_id -> label}. Most commonly this would be the value from
#         Vocabulary.get_index_to_token_vocabulary()
#
#     Returns
#     -------
#     ``List[Tuple[int, int]]``
#         The allowed transitions (from_label_id, to_label_id).
#     """
#     num_labels = len(labels)
#     start_tag = num_labels
#     end_tag = num_labels + 1
#     labels_with_boundaries = list(labels.items()) + [(start_tag, "START"), (end_tag, "END")]
#
#     num_total_labels = len(labels_with_boundaries)
#     total_labels = [v for k, v in labels_with_boundaries]
#
#     allowed_matrix = pd.DataFrame(
#         data=np.zeros([num_total_labels, num_total_labels]),
#         index=total_labels,
#         columns=total_labels,
#         dtype=np.bool
#     )
#
#     allowed = []
#     for from_label_index, from_label in labels_with_boundaries:
#         from_tag, from_entity = extract_tag_entity(from_label)
#
#         for to_label_index, to_label in labels_with_boundaries:
#             to_tag, to_entity = extract_tag_entity(to_label)
#
#             if is_transition_allowed(constraint_type, from_tag, from_entity,
#                                      to_tag, to_entity):
#                 allowed.append((from_label_index, to_label_index))
#                 allowed_matrix.at[from_label, to_label] = True
#     return allowed, allowed_matrix


def allowed_transitions(constraint_type: str, labels: Dict[int, str], tag_set: Dict[int, str] = None) -> (List[Tuple[int, int]], pd.DataFrame):
    """
    Given labels and a constraint type, returns the allowed transitions. It will
    additionally include transitions for the start and end states, which are used
    by the conditional random field.

    Parameters
    ----------
    constraint_type : ``str``, required
        Indicates which constraint to apply. Current choices are
        "BIO", "IOB1", "BIOUL", and "BMES".
    labels : ``Dict[int, str]``, required
        A mapping {label_id -> label}. Most commonly this would be the value from
        Vocabulary.get_index_to_token_vocabulary()

    Returns
    -------
    ``List[Tuple[int, int]]``
        The allowed transitions (from_label_id, to_label_id).
    """
    if not tag_set:
        tag_set = labels

    num_labels = len(tag_set)
    start_tag = num_labels
    end_tag = num_labels + 1
    labels_with_boundaries = list(labels.items()) + [(start_tag, "START"), (end_tag, "END")]

    tag_set_with_boundaries = list(tag_set.items()) + [(start_tag, "START"), (end_tag, "END")]

    num_total_labels = len(tag_set_with_boundaries)
    total_labels = [v for k, v in tag_set_with_boundaries]

    allowed_matrix = pd.DataFrame(
        data=np.zeros([num_total_labels, num_total_labels]),
        index=total_labels,
        columns=total_labels,
        dtype=np.bool
    )

    allowed = []
    for from_label_index, from_label in labels_with_boundaries:
        from_tag, from_entity = extract_tag_entity(from_label)

        for to_label_index, to_label in labels_with_boundaries:
            to_tag, to_entity = extract_tag_entity(to_label)

            if is_transition_allowed(constraint_type, from_tag, from_entity,
                                     to_tag, to_entity):
                allowed.append((from_label_index, to_label_index))
                allowed_matrix.iloc[from_label_index, to_label_index] = True
    return allowed, allowed_matrix


def is_transition_allowed(constraint_type: str,
                          from_tag: str,
                          from_entity: str,
                          to_tag: str,
                          to_entity: str):
    """
    Given a constraint type and strings ``from_tag`` and ``to_tag`` that
    represent the origin and destination of the transition, return whether
    the transition is allowed under the given constraint type.

    Parameters
    ----------
    constraint_type : ``str``, required
        Indicates which constraint to apply. Current choices are
        "BIO", "IOB1", "BIOUL", and "BMES".
    from_tag : ``str``, required
        The tag that the transition originates from. For example, if the
        label is ``I-PER``, the ``from_tag`` is ``I``.
    from_entity: ``str``, required
        The entity corresponding to the ``from_tag``. For example, if the
        label is ``I-PER``, the ``from_entity`` is ``PER``.
    to_tag : ``str``, required
        The tag that the transition leads to. For example, if the
        label is ``I-PER``, the ``to_tag`` is ``I``.
    to_entity: ``str``, required
        The entity corresponding to the ``to_tag``. For example, if the
        label is ``I-PER``, the ``to_entity`` is ``PER``.

    Returns
    -------
    ``bool``
        Whether the transition is allowed under the given ``constraint_type``.
    """
    # pylint: disable=too-many-return-statements
    if to_tag == "START" or from_tag == "END":
        # Cannot transition into START or from END
        return False

    if constraint_type == "BIOUL":
        if from_tag == "START":
            return to_tag in ('O', 'B', 'U')
        if to_tag == "END":
            return from_tag in ('O', 'L', 'U')
        return any([
                # O can transition to O, B-* or U-*
                # L-x can transition to O, B-*, or U-*
                # U-x can transition to O, B-*, or U-*
                from_tag in ('O', 'L', 'U') and to_tag in ('O', 'B', 'U'),
                # B-x can only transition to I-x or L-x
                # I-x can only transition to I-x or L-x
                from_tag in ('B', 'I') and to_tag in ('I', 'L') and from_entity == to_entity
        ])
    elif constraint_type == "BIO":
        if from_tag == "START":
            return to_tag in ('O', 'B')
        if to_tag == "END":
            return from_tag in ('O', 'B', 'I')
        return any([
                # Can always transition to O or B-x
                to_tag in ('O', 'B'),
                # Can only transition to I-x from B-x or I-x
                to_tag == 'I' and from_tag in ('B', 'I') and from_entity == to_entity
        ])
    elif constraint_type == "IOB1":
        if from_tag == "START":
            return to_tag in ('O', 'I')
        if to_tag == "END":
            return from_tag in ('O', 'B', 'I')
        return any([
                # Can always transition to O or I-x
                to_tag in ('O', 'I'),
                # Can only transition to B-x from B-x or I-x, where
                # x is the same tag.
                to_tag == 'B' and from_tag in ('B', 'I') and from_entity == to_entity
        ])
    elif constraint_type == "BMES":
        if from_tag == "START":
            return to_tag in ('B', 'S')
        if to_tag == "END":
            return from_tag in ('E', 'S')
        return any([
                # Can only transition to B or S from E or S.
                to_tag in ('B', 'S') and from_tag in ('E', 'S'),
                # Can only transition to M-x from B-x, where
                # x is the same tag.
                to_tag == 'M' and from_tag in ('B', 'M') and from_entity == to_entity,
                # Can only transition to E-x from B-x or M-x, where
                # x is the same tag.
                to_tag == 'E' and from_tag in ('B', 'M') and from_entity == to_entity,
        ])
    else:
        raise ConfigurationError(f"Unknown constraint type: {constraint_type}")


from setuptools import setup, find_packages

setup(
    name="tf_crf_layer",
    version="0.2.3",
    packages=find_packages(include=["tf_crf_layer", "tf_crf_layer.*"]),
    url="https://github.com/howl-anderson/tf_crf_layer",
    license="MIT",
    install_requires=["tensorflow", "tokenizer_tools"],
    tests_require=["pytest", "numpy", "keras", "pandas"],
    author="Xiaoquan Kong",
    author_email="u1mail2me@gmail.com",
    description="CRF layer for TensorFlow 1.x",
)



import tensorflow as tf
from tensorflow.python.keras.metrics import Mean
from tensorflow.python.keras import backend as K

# for future reference:
# B: batch size
# M: intent number
# n: where n is the tag set number
# n+2: n plus start and end tag
# N: n * n
# T: sequence length
# F: feature number
from tf_crf_layer.keras_utils import register_keras_custom_object


@register_keras_custom_object
class SequenceCorrectness(Mean):
    def __init__(self, name="sequence_correctness", dtype=None):
        """Creates a `CategoricalAccuracy` instance.
        Args:
          name: (Optional) string name of the metric instance.
          dtype: (Optional) data type of the metric result.
        """
        super(SequenceCorrectness, self).__init__(
            name=name,
            dtype=dtype
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        """

        :param y_true: Tensor shape: (B, T)
        :param y_pred: Tensor shape (B, T)
        :param sample_weight: None or Tensor shape (B, T)
        :return:
        """
        judge = K.equal(y_pred, y_true)  # shape: (B, T)
        value = K.all(judge, axis=1, keepdims=False)  # shape: (B, )

        # # # DEBUG: output training value
        # print_op = tf.print(value)
        # with tf.control_dependencies([print_op]):
        #     value = tf.identity(value)

        super(SequenceCorrectness, self).update_state(value)



import tensorflow as tf
from tensorflow.python.keras import backend as K
from tensorflow.python.keras.losses import sparse_categorical_crossentropy, \
    categorical_crossentropy
from tensorflow.python.keras.utils.losses_utils import ReductionV2

from tf_crf_layer import keras_utils
from tf_crf_layer.keras_utils import register_keras_custom_object


def crf_nll(y_true, y_pred):
    crf, idx = y_pred._keras_history[:2]
#     if crf._outbound_nodes:
#         raise TypeError('When learn_model="join", CRF must be the last layer.')
#     if crf.sparse_target:
#         y_true = tf.one_hot(tf.cast(y_true[:, :, 0], 'int32'), crf.units)

    # node = crf._inbound_nodes[idx]
    #
    # print(node)
    # print(node.input_tensors)

    # X = node.input_tensors[0]
    # mask = node.input_tensors[1]

    # nloglik = crf.get_negative_log_likelihood(y_pred, y_true, mask)
    nloglik = crf.get_negative_log_likelihood(y_true)  # shape: (batch_size, )

    return nloglik


@register_keras_custom_object
def crf_loss(y_true, y_pred):
    crf, idx = y_pred._keras_history[:2]
    print(crf, idx)

    if crf.learn_mode == 'join':
        return crf_nll(y_true, y_pred)
    else:
        if crf.sparse_target:
            return sparse_categorical_crossentropy(y_true, y_pred)
        else:
            return categorical_crossentropy(y_true, y_pred)


# @register_keras_custom_object
# class ConditionalRandomFieldLoss(tf.keras.losses.Loss):
#     def __init__(self, name=None):
#         super(ConditionalRandomFieldLoss, self).__init__(reduction=ReductionV2.SUM_OVER_BATCH_SIZE, name=name)
#
#     def get_config(self):
#         config = {}
#         base_config = super(ConditionalRandomFieldLoss, self).get_config()
#         return dict(list(base_config.items()) + list(config.items()))
#
#     def call(self, y_true, y_pred):
#         loss_vector = crf_loss(y_true, y_pred)
#
#         two_dim_loss = K.expand_dims(loss_vector)
#
#         mask = K.ones_like(y_true)
#
#         full_shape_loss = tf.math.multiply(two_dim_loss, mask)
#
#         return full_shape_loss


@register_keras_custom_object
class ConditionalRandomFieldLoss(object):
    def get_config(self):
        return {}

    def __call__(self, y_true, y_pred, sample_weight=None):
        loss_vector = crf_loss(y_true, y_pred)

        return K.mean(loss_vector)



import numpy as np
import tensorflow as tf
from tensorflow.python.keras import backend as K
from tensorflow.python.keras.utils import tf_utils
from tensorflow.python.keras import initializers, regularizers, constraints, activations
from tensorflow.python.keras.layers import InputSpec, Layer
from tensorflow.contrib.crf import crf_log_likelihood
from tf_crf_layer.crf import crf_decode

from tf_crf_layer import keras_utils

"""
TODO

* learn_mode is not supported
* test_mode is not supported
* sparse_target is not supported
* use_boundary need test
* input_dim is not know how to use
* unroll is not supported

* left padding of mask is not supported

* not test yet if CRF is the first layer
"""


# for future reference:
# B: batch size
# M: intent number
# n: where n is the tag set number
# n+2: n plus start and end tag
# N: n * n
# T: sequence length
# F: feature number


def add_pooling_strategies(x):
    # x: shape (B, M, n+2, n+2)
    # return (B, n+2, n+2)
    return K.sum(x, axis=1, keepdims=False)


pooling_strategies = {"add": add_pooling_strategies}


@keras_utils.register_keras_custom_object
class CRF(Layer):
    def __init__(
        self,
        units,
        learn_mode="join",
        test_mode=None,
        sparse_target=False,
        use_boundary=False,
        use_bias=True,
        use_kernel=True,
        activation="linear",
        kernel_initializer="glorot_uniform",
        chain_initializer="orthogonal",
        bias_initializer="zeros",
        left_boundary_initializer="zeros",
        right_boundary_initializer="zeros",
        kernel_regularizer=None,
        chain_regularizer=None,
        boundary_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        chain_constraint=None,
        boundary_constraint=None,
        bias_constraint=None,
        input_dim=None,
        unroll=False,
        transition_constraint=None,
        transition_constraint_matrix=None,  # shape: (M, n+2, n+2)
        pooling_strategy=add_pooling_strategies,
        **kwargs
    ):
        super(CRF, self).__init__(**kwargs)

        # setup mask supporting flag, used by base class (the Layer)
        self.supports_masking = True

        self.units = units  # numbers of tags

        self.learn_mode = learn_mode
        assert self.learn_mode in ["join", "marginal"]

        self.test_mode = test_mode
        if self.test_mode is None:
            self.test_mode = "viterbi" if self.learn_mode == "join" else "marginal"
        else:
            assert self.test_mode in ["viterbi", "marginal"]
        self.sparse_target = sparse_target
        self.use_boundary = use_boundary
        self.use_bias = use_bias
        self.use_kernel = use_kernel

        self.activation = activations.get(activation)

        self.kernel_initializer = initializers.get(kernel_initializer)
        self.chain_initializer = initializers.get(chain_initializer)
        self.left_boundary_initializer = initializers.get(left_boundary_initializer)
        self.right_boundary_initializer = initializers.get(right_boundary_initializer)
        self.bias_initializer = initializers.get(bias_initializer)

        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.chain_regularizer = regularizers.get(chain_regularizer)
        self.boundary_regularizer = regularizers.get(boundary_regularizer)
        self.bias_regularizer = regularizers.get(bias_regularizer)

        self.kernel_constraint = constraints.get(kernel_constraint)
        self.chain_constraint = constraints.get(chain_constraint)
        self.boundary_constraint = constraints.get(boundary_constraint)
        self.bias_constraint = constraints.get(bias_constraint)

        self.input_dim = input_dim
        self.unroll = unroll
        self.transition_constraint = transition_constraint
        self.transition_constraint_matrix = transition_constraint_matrix
        self.pooling_strategy = pooling_strategy

        # value remembered for loss/metrics function
        self.logits = None
        self.nwords = None
        self.mask = None

        # global variable
        self.kernel = None
        self.chain_kernel = None  # shape: (self.units, self.units)
        self.bias = None
        self.left_boundary = None
        self.right_boundary = None
        self.transition_constraint_mask = None
        self.transition_constraint_matrix_mask = None
        self.dynamic_transition_constraint = None
        self.is_training = False

        # for debug
        self.dynamic_transition_constraint_input = None
        self.inputs = None
        self.actual_left_boundary = None
        self.actual_right_boundary = None

    def build(self, input_shape):
        if isinstance(input_shape, list):
            input_shape = input_shape[0]

        input_shape = tuple(tf.TensorShape(input_shape).as_list())
        self.input_spec = [InputSpec(shape=input_shape)]
        self.input_dim = input_shape[-1]

        self.transition_constraint_mask = self.add_weight(
            shape=(self.units + 2, self.units + 2),
            name="transition_constraint_mask",
            initializer=initializers.Constant(self.get_transition_constraint_mask()),
            trainable=False,
        )

        if self.transition_constraint_matrix is not None:
            self.transition_constraint_matrix_mask = self.add_weight(
                shape=self.transition_constraint_matrix.shape,
                name="transition_constraint_mask",
                initializer=initializers.Constant(self.transition_constraint_matrix),
                trainable=False,
            )

        if self.use_kernel:
            # weights that mapping arbitrary tensor to correct shape
            self.kernel = self.add_weight(
                shape=(self.input_dim, self.units),
                name="kernel",
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
            )

        # weights that work as transfer probability of each tags
        self.chain_kernel = self.add_weight(
            shape=(self.units, self.units),
            name="chain_kernel",
            initializer=self.chain_initializer,
            regularizer=self.chain_regularizer,
            constraint=self.chain_constraint,
        )

        # bias that works with self.kernel
        if self.use_kernel and self.use_bias:
            self.bias = self.add_weight(
                shape=(self.units,),
                name="bias",
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
            )
        else:
            self.bias = 0

        # weight of <START> to tag probability and tag to <END> probability
        if self.use_boundary:
            self.left_boundary = self.add_weight(
                shape=(self.units,),
                name="left_boundary",
                initializer=self.left_boundary_initializer,
                regularizer=self.boundary_regularizer,
                constraint=self.boundary_constraint,
            )
            self.right_boundary = self.add_weight(
                shape=(self.units,),
                name="right_boundary",
                initializer=self.right_boundary_initializer,
                regularizer=self.boundary_regularizer,
                constraint=self.boundary_constraint,
            )

        # or directly call self.built = True
        super(CRF, self).build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        # inputs: Tensor(shape(B, T, F)) or (Tensor(B, T, F), Tensor(B, M))
        # mask: Tensor(shape=(B, T), dtype=bool) or None

        training = kwargs.get("training", None)
        if training is None:
            training = K.learning_phase()

        # session = K.get_session()
        #
        # print("training_flag:", session.run(training))

        # # DEBUG: output training value
        # training = tf.Print(training, [training])

        self.is_training = training

        dynamic_transition_constraint_indicator = None

        if isinstance(inputs, list) and len(inputs) > 1:
            assert len(inputs) == 2, "Input must have two input tensors"

            dynamic_transition_constraint_indicator = inputs[1]
            self.dynamic_transition_constraint_input = (
                dynamic_transition_constraint_indicator
            )

            inputs = inputs[0]
            self.inputs = inputs
        else:
            inputs = inputs[0]
            self.inputs = inputs

        if isinstance(mask, list):
            print(mask)
            mask = mask[0]

        if mask is not None:
            assert K.ndim(mask) == 2, "Input mask to CRF must have dim 2 if not None"

        # remember this value for later use
        self.mask = mask

        if dynamic_transition_constraint_indicator is not None:
            # reshape from (B, M) to (B, M, 1, 1)
            constraint_indicator = K.expand_dims(
                K.expand_dims(dynamic_transition_constraint_indicator)
            )

            # shape: (B, M, n+2, n+2)
            raw_constrain_pool = (
                self.transition_constraint_matrix * constraint_indicator
            )

            # shape: (B, n+2, n+2)
            unrolled_dynamic_transition_constraint = self.pooling_strategy(
                raw_constrain_pool
            )

            # shape: (B, n+2, n+2), +2 for start and end tag
            self.dynamic_transition_constraint = unrolled_dynamic_transition_constraint

        logits = self._dense_layer(inputs)
        print("logits:", logits)

        # appending boundary probability info
        if self.use_boundary:
            logits = self.add_boundary_energy(logits, mask)

        print("logits:", logits)

        # remember this value for later use
        self.logits = logits

        nwords = self._get_nwords(inputs, mask)
        print("nwords: {}".format(nwords))

        # remember this value for later use
        self.nwords = nwords

        if self.test_mode == "viterbi":
            test_output = self.get_viterbi_decoding(logits, nwords)
        else:
            # TODO: not supported yet
            pass
            # test_output = self.get_marginal_prob(input, mask)

        if self.learn_mode == "join":
            # WHY: don't remove this line, useless but remote it will cause bug
            test_output = tf.cast(test_output, tf.float32)
            out = test_output
        else:
            # TODO: not supported yet
            pass
            # if self.test_mode == 'viterbi':
            #     train_output = self.get_marginal_prob(input, mask)
            #     out = K.in_train_phase(train_output,
            #                                           test_output)
            # else:
            #     out = test_output

        return out

    def _get_nwords(self, input, mask):
        if mask is not None:
            int_mask = K.cast(mask, tf.int8)
            nwords = self.mask_to_nwords(int_mask)
        else:
            # make a mask tensor from input, then used to generate nwords
            input_energy_shape = tf.shape(input)
            raw_input_shape = tf.slice(input_energy_shape, [0], [2])
            alt_mask = tf.ones(raw_input_shape)

            nwords = self.mask_to_nwords(alt_mask)

        print("nwords: {}".format(nwords))
        return nwords

    def mask_to_nwords(self, mask):
        nwords = K.cast(K.sum(mask, 1), tf.int64)
        return nwords

    @staticmethod
    def shift_left(x, offset=1):
        assert offset > 0
        return K.concatenate([x[:, offset:], K.zeros_like(x[:, :offset])], axis=1)

    @staticmethod
    def shift_right(x, offset=1):
        assert offset > 0
        return K.concatenate([K.zeros_like(x[:, :offset]), x[:, :-offset]], axis=1)

    def compute_energy(self, start, end, energy, mask):
        self.actual_left_boundary = start
        self.actual_right_boundary = end

        if mask is None:
            energy = K.concatenate([energy[:, :1, :] + start, energy[:, 1:, :]], axis=1)

            energy = K.concatenate([energy[:, :-1, :], energy[:, -1:, :] + end], axis=1)
        else:
            mask = K.expand_dims(K.cast(mask, K.floatx()), axis=-1)
            start_mask = K.cast(K.greater(mask, self.shift_right(mask)), K.floatx())

            # original code:
            # end_mask = K.cast(K.greater(self.shift_left(mask), mask), K.floatx())
            # TODO: is this a bug? should be K.greater(mask, self.shift_left(mask)) ?
            # patch applied
            end_mask = K.cast(K.greater(mask, self.shift_left(mask)), K.floatx())
            energy = energy + start_mask * start
            energy = energy + end_mask * end

        return energy

    def compute_boundary_with_constraint(self):
        if self.transition_constraint:
            start_tag = self.units
            end_tag = self.units + 1

            left_boundary = self.left_boundary * self.transition_constraint_mask[
                start_tag, : self.units
            ] + -10000.0 * (
                1 - self.transition_constraint_mask[start_tag, : self.units]
            )
            right_boundary = self.right_boundary * self.transition_constraint_mask[
                : self.units, end_tag
            ] + -10000.0 * (1 - self.transition_constraint_mask[: self.units, end_tag])

            return left_boundary, right_boundary

        if self.transition_constraint_matrix is not None:
            start_tag = self.units
            end_tag = self.units + 1

            # shape: (B, 1, n)
            # left_dynamic_constraint = self.dynamic_transition_constraint[:, start_tag, :self.units]  # shape: (B, n)
            left_dynamic_constraint = tf.slice(
                self.dynamic_transition_constraint,
                [0, start_tag, 0],
                [-1, 1, self.units],
            )  # shape: (B, 1, n)

            # shape: (B, 1, n)
            left_boundary = self.left_boundary * left_dynamic_constraint + -10000.0 * (
                1 - left_dynamic_constraint
            )

            # shape: (B, 1, n)
            # right_dynamic_constraint = self.dynamic_transition_constraint[:, :self.units, end_tag]  # shape: (B, n)
            raw_right_dynamic_constraint = tf.slice(
                self.dynamic_transition_constraint, [0, 0, end_tag], [-1, self.units, 1]
            )  # shape: (B, n, 1)
            right_dynamic_constraint = tf.transpose(
                raw_right_dynamic_constraint, perm=[0, 2, 1]
            )  # shape: (B, 1, n)

            # shape: (B, 1, n)
            right_boundary = (
                self.right_boundary * right_dynamic_constraint
                + -10000.0 * (1 - right_dynamic_constraint)
            )

            return left_boundary, right_boundary

        # no constraint at all, return ordinary energy
        return self.compute_boundary_ordinary()

    def compute_boundary_without_constraint(self):
        return self.compute_boundary_ordinary()

    def compute_boundary_ordinary(self):
        # no any constraint at all

        def expend_scalar_to_3d(x):
            # expend tensor from shape (x, ) to (1, 1, x)
            return K.expand_dims(K.expand_dims(x, 0), 0)

        # shape: (1, 1, n)
        left_boundary = expend_scalar_to_3d(self.left_boundary)

        # shape: (1, 1, n)
        right_boundary = expend_scalar_to_3d(self.right_boundary)

        return left_boundary, right_boundary

    def add_boundary_energy(self, energy, mask):
        """

        :param energy: Tensor(shape=(B, T, F))
        :param mask: Tensor(shape=(B, T), dtype=bool) or None
        :return:
        """

        def compute_constrained_energy():
            left_boundary, right_boundary = self.compute_boundary_with_constraint()

            return self.compute_energy(left_boundary, right_boundary, energy, mask)

        def compute_unconstrained_energy():
            left_boundary, right_boundary = self.compute_boundary_without_constraint()

            return self.compute_energy(left_boundary, right_boundary, energy, mask)

        computed_energy = tf_utils.smart_cond(
            self.is_training, compute_unconstrained_energy, compute_constrained_energy
        )

        return computed_energy

    def compute_static_constrained_chain_kernel(self):
        if self.transition_constraint:
            chain_kernel_mask = self.transition_constraint_mask[
                : self.units, : self.units
            ]
            constrained_transitions = (
                self.chain_kernel * chain_kernel_mask
                + -10000.0 * (1 - chain_kernel_mask)
            )

            return constrained_transitions

    def compute_dynamic_constrained_chain_kernel(self):
        if self.transition_constraint_matrix is not None:
            dynamic_chain_kernel_mask = self.dynamic_transition_constraint[
                :, : self.units, : self.units
            ]
            dynamic_constrained_transitions = (
                self.chain_kernel * dynamic_chain_kernel_mask
                + -10000.0 * (1 - dynamic_chain_kernel_mask)
            )

            return dynamic_constrained_transitions

    def get_viterbi_decoding(self, input_energy, nwords):
        def compute_constrained_chain_kernel():
            dynamic_constrained_transitions = None

            if self.transition_constraint:
                dynamic_constrained_transitions = (
                    self.compute_static_constrained_chain_kernel()
                )

            if self.transition_constraint_matrix is not None:
                dynamic_constrained_transitions = (
                    self.compute_dynamic_constrained_chain_kernel()
                )

            if dynamic_constrained_transitions is None:
                # in case there are no constraint at all
                dynamic_constrained_transitions = self.chain_kernel

            return crf_decode(input_energy, dynamic_constrained_transitions, nwords)

        def compute_unconstrained_chain_kernel():
            return crf_decode(input_energy, self.chain_kernel, nwords)

        pred_ids, _ = tf_utils.smart_cond(
            self.is_training,
            compute_unconstrained_chain_kernel,
            compute_constrained_chain_kernel,
        )

        return pred_ids

    def get_config(self):
        # will be used for loading model from disk,
        # see https://github.com/keras-team/keras/issues/4871#issuecomment-269714512

        config = {
            "units": self.units,
            "learn_mode": self.learn_mode,
            "test_mode": self.test_mode,
            "use_boundary": self.use_boundary,
            "use_bias": self.use_bias,
            "use_kernel": self.use_kernel,
            "sparse_target": self.sparse_target,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "chain_initializer": initializers.serialize(self.chain_initializer),
            "left_boundary_initializer": initializers.serialize(
                self.left_boundary_initializer
            ),
            "right_boundary_initializer": initializers.serialize(
                self.right_boundary_initializer
            ),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "activation": activations.serialize(self.activation),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "chain_regularizer": regularizers.serialize(self.chain_regularizer),
            "boundary_regularizer": regularizers.serialize(self.boundary_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "chain_constraint": constraints.serialize(self.chain_constraint),
            "boundary_constraint": constraints.serialize(self.boundary_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "input_dim": self.input_dim,
            "unroll": self.unroll,
            "transition_constraint": self.transition_constraint,
            "transition_constraint_matrix": self.transition_constraint_matrix,
            "pooling_strategy": self.pooling_strategy,
        }
        base_config = super(CRF, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        output_shape = input_shape[:2]
        return output_shape

    # def compute_mask(self, input, mask=None):
    #     if isinstance(mask, list):
    #         mask = mask[0]
    #
    #     if mask is not None and self.learn_mode == "join":
    #         # transform mask from shape (?, ?) to (?, )
    #         new_mask = K.any(mask, axis=1)
    #         return new_mask
    #
    #     return mask

    # def get_decode_result(self, logits, mask):
    #     nwords = K.cast(K.sum(mask, 1), tf.int64)
    #
    #     pred_ids, _ = crf_decode(logits, self.chain_kernel,
    #                                             nwords)
    #
    #     return pred_ids

    def get_negative_log_likelihood(self, y_true):
        y_preds = self.logits

        nwords = self.nwords

        y_preds = K.cast(y_preds, tf.float32)
        y_true = K.cast(y_true, tf.int64)
        nwords = K.cast(nwords, tf.int32)
        self.chain_kernel = K.cast(self.chain_kernel, tf.float32)

        # print_op = tf.print([tf.shape(y_preds), tf.shape(y_true), tf.shape(nwords)])
        # with tf.control_dependencies([print_op]):
        #     y_preds = tf.identity(y_preds)

        # print(K.get_session().graph.get_name_scope())

        log_likelihood, _ = crf_log_likelihood(
            y_preds, y_true, nwords, self.chain_kernel
        )

        return -log_likelihood

    def get_accuracy(self, y_true, y_pred):
        judge = K.cast(K.equal(y_pred, y_true), K.floatx())
        if self.mask is None:
            result = K.mean(judge)
            return result
        else:
            mask = K.cast(self.mask, K.floatx())
            result = K.sum(judge * mask) / K.sum(mask)
            return result

    def _dense_layer(self, input_):
        """
        mapping any number of features to votes for each tags

        :param input_: Tensor(shape=(B, T, F))
        :return: Tensor(shape=(B, T, F))
        """
        # TODO: can simply just use tf.keras.layers.dense ?
        if self.use_kernel:
            return self.activation(K.dot(input_, self.kernel) + self.bias)

        return input_

    def get_transition_constraint_mask(self):
        if not self.transition_constraint:
            # All transitions are valid.
            constraint_mask = np.ones([self.units + 2, self.units + 2])
        else:
            constraint_mask = np.zeros([self.units + 2, self.units + 2])
            for i, j in self.transition_constraint:
                constraint_mask[i, j] = 1.0

        return constraint_mask

    def compute_effective_boundary(self, energy):
        """

        :return: Tensor(shape=(B, 1, n)) or Tensor(shape=(1, 1, n))
        """

        def compute_constrained_boundary():
            if self.transition_constraint:
                start_tag = self.units
                end_tag = self.units + 1

                left_boundary = self.left_boundary * self.transition_constraint_mask[
                    start_tag, : self.units
                ] + -10000.0 * (
                    1 - self.transition_constraint_mask[start_tag, : self.units]
                )
                right_boundary = self.right_boundary * self.transition_constraint_mask[
                    : self.units, end_tag
                ] + -10000.0 * (
                    1 - self.transition_constraint_mask[: self.units, end_tag]
                )

                return left_boundary, right_boundary

            if self.transition_constraint_matrix is not None:
                start_tag = self.units
                end_tag = self.units + 1

                # shape: (B, 1, n)
                # left_dynamic_constraint = self.dynamic_transition_constraint[:, start_tag, :self.units]  # shape: (B, n)
                left_dynamic_constraint = tf.slice(
                    self.dynamic_transition_constraint,
                    [0, start_tag, 0],
                    [-1, 1, self.units],
                )  # shape: (B, 1, n)

                # shape: (B, 1, n)
                left_boundary = (
                    self.left_boundary * left_dynamic_constraint
                    + -10000.0 * (1 - left_dynamic_constraint)
                )

                # shape: (B, 1, n)
                # right_dynamic_constraint = self.dynamic_transition_constraint[:, :self.units, end_tag]  # shape: (B, n)
                raw_right_dynamic_constraint = tf.slice(
                    self.dynamic_transition_constraint,
                    [0, 0, end_tag],
                    [-1, self.units, 1],
                )  # shape: (B, n, 1)
                right_dynamic_constraint = tf.transpose(
                    raw_right_dynamic_constraint, perm=[0, 2, 1]
                )  # shape: (B, 1, n)

                # shape: (B, 1, n)
                right_boundary = (
                    self.right_boundary * right_dynamic_constraint
                    + -10000.0 * (1 - right_dynamic_constraint)
                )

                return left_boundary, right_boundary

            return compute_unconstrained_boundary()

        def compute_unconstrained_boundary():
            # no any constraint at all

            def expend_scalar_to_3d(x):
                # expend tensor from shape (x, ) to (1, 1, x)
                return K.expand_dims(K.expand_dims(x, 0), 0)

            # shape: (1, 1, n)
            left_boundary = expend_scalar_to_3d(self.left_boundary)

            # shape: (1, 1, n)
            right_boundary = expend_scalar_to_3d(self.right_boundary)

            return left_boundary, right_boundary

        left_boundary, right_boundary = tf_utils.smart_cond(
            self.is_training,
            compute_unconstrained_boundary,
            compute_constrained_boundary,
        )

        return left_boundary, right_boundary
